import inspect
from contextlib import ExitStack, contextmanager
from typing import Iterator

import numpy as np

from gustavgrad.tensor import Tensor


class Parameter(Tensor):
    " A Parameter is used to register a Tensor as part of a Module "

    def __init__(self, *shape: int) -> None:
        # TODO: Add possibility to use custom initialization schemes
        super().__init__(np.random.randn(*shape), requires_grad=True)


class Module:
    " A Module wraps multiple Parameters in a single computational block "

    def parameters(self) -> Iterator[Parameter]:
        for _, value in inspect.getmembers(self):
            if isinstance(value, Parameter):
                yield value
            elif isinstance(value, Module):
                yield from value.parameters()

    def zero_grad(self) -> None:
        for parameter in self.parameters():
            parameter.zero_grad()

    @contextmanager
    def no_grad(self) -> Iterator[None]:
        with ExitStack() as stack:
            for parameter in self.parameters():
                stack.enter_context(parameter.no_grad())
            yield



""" Loss functions"""
from abc import ABC, abstractmethod

import numpy as np

from gustavgrad.function import _sigmoid
from gustavgrad.tensor import Dependency, Tensor, Tensorable, ensure_tensor


class Loss(ABC):
    @abstractmethod
    def loss(self, target: Tensorable, pred: Tensor) -> Tensor:
        raise NotImplementedError


class SquaredErrorLoss(Loss):
    """ Squared error loss """

    def loss(self, target: Tensorable, pred: Tensor) -> Tensor:
        error = ensure_tensor(target) - pred
        return (error * error).sum()


class LogitBinaryCrossEntropy(Loss):
    """ Binary cross entropy with logits
    Applies the logistic function (sigmoid) to logits before calculating loss
    """

    def loss(self, target: Tensorable, logits: Tensor) -> Tensor:

        tensor_target = ensure_tensor(target)

        # Calculate stable binary cross entropy loss
        x = np.clip(logits.data, 0, None)
        neg_abs = -np.abs(logits.data)
        data = (
            x - logits.data * tensor_target.data + np.log(1 + np.exp(neg_abs))
        )
        # Just do the mean for now
        data = data.mean()

        requires_grad = logits.requires_grad
        depends_on = []
        if requires_grad:

            def grad_fn(grad: np.ndarray) -> np.ndarray:
                return grad * (_sigmoid(logits.data) - tensor_target.data)

            depends_on.append(Dependency(logits, grad_fn))

        return Tensor(data, requires_grad, depends_on)



"""
Tensor Class and affiliated functions
"""
from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np


class Dependency(NamedTuple):
    """
    A dependency consists of a tensor and a gradient function.
    The idea is that if tensor t2 = f(t1), then t2 should have the dependency:
    (t1, df/dt1*dt2/dtf).
    If we have a gradient dJ/dt2, we can now calculate the gradient of t1 as:
    dJ/dt1 = df/dt1*dt2/dtf*dJ/dt2
    """

    tensor: "Tensor"
    grad_fn: Callable[[np.ndarray], np.ndarray]


# Types that can be converted to an ndarray
Arrayable = Union[np.ndarray, list, float]


def ensure_array(data: Arrayable) -> np.ndarray:
    if isinstance(data, np.ndarray):
        return data
    else:
        return np.asarray(data)


# Types that can be converted to a tensor
Tensorable = Union["Tensor", np.ndarray, float]


def ensure_tensor(data: Tensorable) -> "Tensor":
    if isinstance(data, Tensor):
        return data
    else:
        return Tensor(data)


_TensorType = TypeVar("_TensorType", bound="Tensor")


class Tensor:
    """
    A tensor is an ndarray that keeps track of it's own gradient,
    and can propagate it's gradient backward to tensors it depends on.

    For example, if t3 = t1 + t2, then the tensor t3 depends on t1 and t2.
    """

    def __init__(
        self,
        data: Arrayable,
        requires_grad: bool = False,
        depends_on: List[Dependency] = None,
    ) -> None:

        if depends_on is None:
            depends_on = []

        # Private, see self.data() property for explanation
        # TODO: Perhaps cast to float?
        self._data = ensure_array(data)
        self._requires_grad = requires_grad
        self.depends_on = depends_on
        self.grad: Optional["Tensor"] = None
        self.shape: tuple = self._data.shape

        if self.requires_grad:
            self.zero_grad()

    def __repr__(self) -> str:
        return (
            "Tensor(data="
            + ("\n" if self.data.ndim > 1 else "")
            + f"{self.data}, requires_grad={self.requires_grad})"
        )

    @property
    def requires_grad(self) -> bool:
        return self._requires_grad

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, value: np.ndarray) -> None:
        """ Invalidate the gradient of the tensor if data is modified in-place
        """
        if value.shape != self.shape:
            raise RuntimeError("Cannot change shape of Tensor")

        self.grad = None
        self._data = value

    def zero_grad(self) -> None:
        """ Set gradient to 0"""
        self.grad = np.zeros_like(self.data, dtype=np.float)

    def backward(self, grad: np.ndarray = None) -> None:
        """ Propagate a gradient backward to all dependencies"""

        if grad is None:
            if self.shape == ():
                grad = np.asarray(1.0)
            else:
                raise RuntimeError("Must specify gradient for non-zero Tensor")

        self.grad += grad

        for tensor, grad_fn in self.depends_on:
            tensor.backward(grad_fn(grad))

    @contextmanager
    def no_grad(self) -> Iterator[None]:
        original_requires_grad = self.requires_grad
        self._requires_grad = False
        try:
            yield
        finally:
            self._requires_grad = original_requires_grad

    def __add__(self, other: Tensorable) -> "Tensor":
        return _add(self, ensure_tensor(other))

    def __radd__(self, other: Tensorable) -> "Tensor":
        return _add(ensure_tensor(other), self)

    def __iadd__(self: _TensorType, other: Tensorable) -> _TensorType:
        self.data += ensure_tensor(other).data
        return self

    def __sub__(self, other: Tensorable) -> "Tensor":
        return _sub(self, ensure_tensor(other))

    def __rsub__(self, other: Tensorable) -> "Tensor":
        return _sub(ensure_tensor(other), self)

    def __isub__(self: _TensorType, other: Tensorable) -> _TensorType:
        self.data -= ensure_tensor(other).data
        return self

    def __mul__(self, other: Tensorable) -> "Tensor":
        return _mul(self, ensure_tensor(other))

    def __rmul__(self, other: Tensorable) -> "Tensor":
        return _mul(ensure_tensor(other), self)

    def __imul__(self: _TensorType, other: Tensorable) -> _TensorType:
        self.data *= ensure_tensor(other).data
        return self

    def __matmul__(self, other: "Tensor") -> "Tensor":
        return _matmul(self, other)

    def __getitem__(self, idxs: Any) -> "Tensor":
        return _slice(self, idxs)

    def sum(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> "Tensor":
        return _sum(self, axis)


""" Functions on tensors """


def _add(t1: Tensor, t2: Tensor) -> Tensor:
    """ Adds two tensors

    Gradients of t1 and t2 are the same as t3, so the gradient function is the
    identity function, UNLESS some dimension of t1 or t2 was broadcasted in the
    addition, in which case the gradient is increased.
    """

    data = t1.data + t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            return sum_out_broadcasted_dims(grad, t1.shape)

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            return sum_out_broadcasted_dims(grad, t2.shape)

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _sub(t1: Tensor, t2: Tensor) -> Tensor:
    """ SUbtracts tensor t2 from tensor t1
    """

    data = t1.data - t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            return sum_out_broadcasted_dims(grad, t1.shape)

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            """ When backpropagating to t2, inverse the gradient. """
            return sum_out_broadcasted_dims(-grad, t2.shape)

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _mul(t1: Tensor, t2: Tensor) -> Tensor:
    """ Multiplies two tensors element-wise.

    If t3 = t1 * t2
    dJ/dt1 = dJ/t3 * dt3/dt1, dt3/dt1 = t2
    dJ/dt2 = dJ/t3 * dt3/dt2, dt3/dt2 = t1
    BUT broadcasting must be taken into account the same way it was done for
    addition
    """

    data = t1.data * t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            grad = grad * t2.data
            return sum_out_broadcasted_dims(grad, t1.shape)

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            grad = grad * t1.data
            return sum_out_broadcasted_dims(grad, t2.shape)

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _matmul(t1: Tensor, t2: Tensor) -> Tensor:
    """ Matrix-multiplies two tensors element-wise.

    t1: (N,K)
    t2: (K,M)
    t3: (N,M)
    """

    data = t1.data @ t2.data  # (N,M)
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            return grad @ t2.data.T  # (N,K)

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            return t1.data.T @ grad  # (K,M)

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _sum(
    tensor: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None
) -> Tensor:
    """
    Sum a tensor along given axis. The gradient function is simply the identity
    function.
    """
    data = tensor.data.sum(axis)
    requires_grad = tensor.requires_grad
    depends_on = []
    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            if axis is not None:
                # Expand the summed axis to enable correct broadcasting
                grad = np.expand_dims(grad, axis)
            return grad * np.ones_like(tensor.data)

        depends_on.append(Dependency(tensor, grad_fn))

    return Tensor(data, requires_grad, depends_on)


def _slice(tensor: Tensor, idxs: Any) -> Tensor:
    """ Slices a tensor using numpy slicing logic"""
    data = tensor.data[idxs]
    requires_grad = tensor.requires_grad
    depends_on = []
    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            new_grad = np.zeros_like(tensor.data)
            new_grad[
                idxs
            ] = grad  # pylint: disable=unsupported-assignment-operation
            return new_grad

        depends_on.append(Dependency(tensor, grad_fn))

    return Tensor(data, requires_grad, depends_on)


""" Helper Functions """


def sum_out_broadcasted_dims(
    grad: np.ndarray, tensor_shape: tuple
) -> np.ndarray:
    """ Handle Broadcasting:
    1. If t1 is shape(N,) and grad is shape (M, K, N), then the grad of t3 with
    respect to t1 should be summed across the added dimensions.
    2. If t1 has any dimensions with a single entry, eg. (N, 1, K), then that
    dimension will also have been broadcasted if the same deimension is >1 for
    t2. Thus, the gradient needs to be summed across this dimension as well.
    """
    # Sum across added dimensions (Step 1)
    added_dims = grad.ndim - len(tensor_shape)
    for _ in range(added_dims):
        grad = grad.sum(0)
    # Sum across singular dimensions (Step 2)
    for i, dim in enumerate(tensor_shape):
        if dim == 1:
            grad = grad.sum(i, keepdims=True)
    return grad



" Optimizers for gustavgrad Modules "

from gustavgrad.module import Module


class SGD:
    " Stochastic Gradient Descent Optimizer"

    def __init__(self, lr: float = 0.001) -> None:
        self.lr = lr

    def step(self, module: Module) -> None:
        for parameter in module.parameters():
            if parameter.grad is not None:
                parameter -= parameter.grad * self.lr



""" Activation functions """
import numpy as np

from gustavgrad.tensor import Dependency, Tensor


def _sigmoid(x: np.ndarray) -> np.ndarray:
    """ Stable sigmoid"""
    mask = x > 0
    res = np.zeros_like(x)
    res[mask] = 1 / (1 + np.exp(-x[mask]))
    res[~mask] = np.exp(x[~mask]) / (1 + np.exp(x[~mask]))
    return res


def _sigmoid_prime(x: np.ndarray) -> np.ndarray:
    s = _sigmoid(x)
    return s * (1 - s)


def sigmoid(tensor: Tensor) -> Tensor:
    """Applies the logistic function element-wise to tensor"""
    data = _sigmoid(tensor.data)
    requires_grad = tensor.requires_grad
    depends_on = []
    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            return grad * _sigmoid_prime(tensor.data)

        depends_on.append(Dependency(tensor, grad_fn))

    return Tensor(data, requires_grad, depends_on)


def tanh(tensor: Tensor) -> Tensor:
    """Applies hyperbolic tangent element-wise to tensor"""
    data = np.tanh(tensor.data)
    requires_grad = tensor.requires_grad
    depends_on = []
    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            return grad * (1 - np.tanh(tensor.data) ** 2)

        depends_on.append(Dependency(tensor, grad_fn))

    return Tensor(data, requires_grad, depends_on)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gustavgrad']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.0,<2.0.0']

setup_kwargs = {
    'name': 'gustavgrad',
    'version': '0.2.0',
    'description': '',
    'long_description': "# gustavgrad\n[![Tests](https://github.com/gustavgransbo/gustavgrad/workflows/Tests/badge.svg)](https://github.com/gustavgransbo/gustavgrad/actions?workflow=Tests)\n[![codecov](https://codecov.io/gh/gustavgransbo/gustavgrad/branch/master/graph/badge.svg)](https://codecov.io/gh/gustavgransbo/gustavgrad)\n[![PyPI](https://img.shields.io/pypi/v/gustavgrad.svg)](https://pypi.org/project/gustavgrad)\n[![Python Version](https://img.shields.io/pypi/pyversions/gustavgrad.svg)](https://github.com/gustavgransbo/gustavgrad)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nAn autograd library built on NumPy, inspired by [Joel Grus's livecoding](https://github.com/joelgrus/autograd/tree/master).\n\n## Installation\n### With pip\n```bash\npip install gustavgrad\n```\n\n## How to use the library\n### `Tensor` operations\nThe `Tensor` class is the cornerstone of `gustavgrad`.\nIt behaves much like an ordinary `numpy.ndarray`.\n\n`Tensors` can be added together,\n```python\n>>> from gustavgrad import Tensor\n>>> x = Tensor([1, 2, 3])\n>>> x + x\nTensor(data=[2 4 6], requires_grad=False)\n```\n... subracted from each other,\n```python\n>>> x - x\nTensor(data=[0 0 0], requires_grad=False)\n```\n... their dot-product can calculated,\n```python\n>>> x * x\nTensor(data=[1 4 9], requires_grad=False)\n```\n... and they can be multiplied with each other.\n```python\n>>> y = Tensor([[1], [2], [3]])\n>>> x @ y\nTensor(data=[14], requires_grad=False)\n```\n`Tensor` operations also support broadcasting:\n```python\n>>> x * 3\nTensor(data=[3 6 9], requires_grad=False)\n>>> z = Tensor([[1, 2, 3], [4, 5, 6]])\n>>> x * z\nTensor(data=\n[[ 1  4  9]\n [ 4 10 18]], requires_grad=False)\n```\n### Automatic backpropagation\nBut a `Tensor` is not just an `ndarray`, they also keep track of their own gradient.\n```python\n>>> speed = Tensor(1, requires_grad=True)\n>>> time = Tensor(10, requires_grad=True)\n>>> distance = speed * time\n>>> distance\nTensor(data=10, requires_grad=True)\n```\nIf a `Tensor` is created as the result of a `Tensor` operation involving a `Tensor` with `requires_grad=True`,\nthe resulting `Tensor` will be able to backpropagate it's own gradient to it's ancestor.\n```python\n>>> distance.backward()\n>>> speed.grad\narray(10.)\n>>> time.grad\narray(1.)\n```\nBy calling the `backward` method on `distance` the gradient of `speed` and `time` is automatically updated.\nWe can see that increasing `speed` by 1 would result in an increase in `distance` by 10, while an increase\nin `time` by 1 would only increase `distance` by 1.\n\nThe `Tensor` class supports backpropagation over arbitrary compositions of `Tensor` operations.\n```python\n>>> t1 = Tensor([[1, 2, 3], [4, 5, 6]], requires_grad=True)\n>>> t2 = Tensor([[1], [2], [3]])\n>>> t3 = t1 @ t2 + 1\n>>> t4 = t3 * 7\n>>> t5 = t4.sum()\n>>> t5.backward()\n>>> t1.grad\narray([[ 7., 14., 21.],\n       [ 7., 14., 21.]])\n```\n### The `Module` API\n`gustavgrad` provides some tools to simplify setting up and training machine learning models.\nThe `Module` API makes it easier to manage multiple related `Tensors` by registering them as\n`Parameters`.\nA `Parameter` is just a randomly initialized `Tensor`.\n```python\nfrom gustavgrad.module import Module, Parameter\nfrom gustavgrad.function import tanh\nclass MultilayerPerceptron(Module):\n    def __init__(self, input_size: int, output_size: int, hidden_size: int = 100) -> None:\n        self.layer1 = Parameter(input_size, hidden_size)\n        self.bias1 = Parameter(hidden_size)\n        self.layer2 = Parameter(hidden_size, output_size)\n        self.bias2 = Parameter(output_size)\n        \n    def predict(self, x: Tensor) -> Tensor:\n        x = x @ self.layer1 + self.bias1\n        x = tanh(x)\n        x = x @ self.layer2 + self.bias2\n        return x\n```\nBy subclassing `Module` our `MultilayerPerceptron` class automatically gets some helper methods\nfor managing its `Parameters`.\nLet's create a `MultilayerPerceptron` that tries to learn the XOR function.\n```python\nxor_input = Tensor([[0, 0], [0, 1], [1, 0], [1, 1]])\nxor_targets = Tensor([[0], [1], [1], [0]])\nxor_mlp = MultilayerPerceptron(input_size=2, output_size=1, hidden_size=4)\n```\nWe can use the model to make predictions on the `xor_input` `Tensor`.\n```python\n>>> predictions = xor_mlp.predict(xor_input)\n>>> predictions\nTensor(data=\n[[-1.79888385]\n [-1.07965756]\n [ 0.34373135]\n [ 1.63366069]], requires_grad=True)\n```\nThe predictions of the randomly initialized model aren't right, but we can improve the model by\ncalculating the gradient of it's `Parameters` in respect to a loss function.\n```python\nfrom gustavgrad.loss import SquaredErrorLoss\nse_loss = SquaredErrorLoss()\nloss = se_loss.loss(xor_targets, predictions)\nloss.backward()\n```\n`loss` is a `Tensor`, so we can call its `backward` method to do backpropagation through our `xor_mlp`.\nWe can then adjust the weights of all `Parameters` in `xor_mlp` using gradient descent:\n```python\nfrom gustavgrad.optim import SGD\noptim = SGD(lr=0.01)\noptim.step(xor_mlp)\n```\nAfter updating the weights we can reset the gradients of all parameters and make new predictions:\n```python\n>>> xor_mlp.zero_grad()\n>>> predictions = xor_mlp.predict(xor_input)\n>>> predictions\nTensor(data=\n[[-1.51682686]\n [-0.78583272]\n [ 0.55994602]\n [ 1.67962174]], requires_grad=True)\n```\nSee [examples/xor.py](examples/xor.py) for a full example of how `gustavgrad` can be used to learn the XOR function.\nThe [examples](examples/) directory also contains some other basic examples of how the library can be used.\n",
    'author': 'Gustav Gränsbo',
    'author_email': 'gustav.gransbo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gustavgransbo/gustavgrad',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
